# pylint:disable=no-member

from time import sleep

from dagster import (
    Field,
    InputDefinition,
    Int,
    List,
    Output,
    OutputDefinition,
    PresetDefinition,
    pipeline,
    solid,
)
from dagster.core.test_utils import default_mode_def_for_test


@solid(
    input_defs=[InputDefinition("units", List[Int])], output_defs=[OutputDefinition(Int, "total")]
)
def sleeper(context, units):
    tot = 0
    for sec in units:
        context.log.info("Sleeping for {} seconds".format(sec))
        sleep(sec)
        tot += sec

    return tot


@solid(
    config_schema=Field([int], is_required=False, default_value=[1, 1, 1, 1]),
    output_defs=[
        OutputDefinition(List[Int], "out_1"),
        OutputDefinition(List[Int], "out_2"),
        OutputDefinition(List[Int], "out_3"),
        OutputDefinition(List[Int], "out_4"),
    ],
)
def giver(context):
    units = context.solid_config
    queues = [[], [], [], []]
    for i, sec in enumerate(units):
        queues[i % 4].append(sec)

    yield Output(queues[0], "out_1")
    yield Output(queues[1], "out_2")
    yield Output(queues[2], "out_3")
    yield Output(queues[3], "out_4")


@solid(
    input_defs=[
        InputDefinition("in_1", Int),
        InputDefinition("in_2", Int),
        InputDefinition("in_3", Int),
        InputDefinition("in_4", Int),
    ],
    output_defs=[OutputDefinition(Int)],
)
def total(_, in_1, in_2, in_3, in_4):
    return in_1 + in_2 + in_3 + in_4


@pipeline(
    description=(
        "Demo diamond-shaped pipeline that has four-path parallel structure of solids.  Execute "
        "with the `multi` preset to take advantage of multi-process parallelism."
    ),
    preset_defs=[
        PresetDefinition(
            "multi",
            {
                "execution": {"multiprocess": {}},
                "solids": {"giver": {"config": [2, 2, 2, 2]}},
            },
        )
    ],
    mode_defs=[default_mode_def_for_test],
)
def sleepy_pipeline():
    giver_res = giver()

    total(
        in_1=sleeper(units=giver_res.out_1),
        in_2=sleeper(units=giver_res.out_2),
        in_3=sleeper(units=giver_res.out_3),
        in_4=sleeper(units=giver_res.out_4),
    )



import json
from datetime import datetime

from dagster.core.telemetry import log_action


def log_dagit_telemetry_event(graphene_info, action, client_time, metadata):
    from ..schema.roots.mutation import GrapheneLogTelemetrySuccess

    instance = graphene_info.context.instance
    metadata = json.loads(metadata)
    client_time = datetime.utcfromtimestamp(int(client_time) / 1000)
    log_action(
        instance=instance,
        action=action,
        client_time=client_time,
        elapsed_time=None,
        metadata=metadata,
    )
    return GrapheneLogTelemetrySuccess(action=action)



import graphene

from dagster.core.definitions.events import AssetKey
from dagster.core.workspace.permissions import Permissions

from ...implementation.execution import (
    cancel_partition_backfill,
    create_and_launch_partition_backfill,
    delete_pipeline_run,
    launch_pipeline_execution,
    launch_pipeline_reexecution,
    resume_partition_backfill,
    terminate_pipeline_execution,
    wipe_assets,
)
from ...implementation.external import fetch_workspace, get_full_external_pipeline_or_raise
from ...implementation.telemetry import log_dagit_telemetry_event
from ...implementation.utils import (
    ExecutionMetadata,
    ExecutionParams,
    UserFacingGraphQLError,
    capture_error,
    check_permission,
    pipeline_selector_from_graphql,
)
from ..asset_key import GrapheneAssetKey
from ..backfill import (
    GrapheneCancelBackfillResult,
    GrapheneLaunchBackfillResult,
    GrapheneResumeBackfillResult,
)
from ..errors import (
    GrapheneAssetNotFoundError,
    GrapheneConflictingExecutionParamsError,
    GraphenePresetNotFoundError,
    GraphenePythonError,
    GrapheneReloadNotSupported,
    GrapheneRepositoryLocationNotFound,
    GrapheneRunNotFoundError,
    GrapheneUnauthorizedError,
)
from ..external import GrapheneWorkspace, GrapheneWorkspaceLocationEntry
from ..inputs import GrapheneAssetKeyInput, GrapheneExecutionParams, GrapheneLaunchBackfillParams
from ..pipelines.pipeline import GrapheneRun
from ..runs import (
    GrapheneLaunchRunReexecutionResult,
    GrapheneLaunchRunResult,
    parse_run_config_input,
)
from ..schedules import GrapheneStartScheduleMutation, GrapheneStopRunningScheduleMutation
from ..sensors import (
    GrapheneSetSensorCursorMutation,
    GrapheneStartSensorMutation,
    GrapheneStopSensorMutation,
)
from ..util import non_null_list


def create_execution_params(graphene_info, graphql_execution_params):
    preset_name = graphql_execution_params.get("preset")
    selector = pipeline_selector_from_graphql(graphql_execution_params["selector"])
    if preset_name:
        if graphql_execution_params.get("runConfigData"):
            raise UserFacingGraphQLError(
                GrapheneConflictingExecutionParamsError(conflicting_param="runConfigData")
            )

        if graphql_execution_params.get("mode"):
            raise UserFacingGraphQLError(
                GrapheneConflictingExecutionParamsError(conflicting_param="mode")
            )

        if selector.solid_selection:
            raise UserFacingGraphQLError(
                GrapheneConflictingExecutionParamsError(
                    conflicting_param="selector.solid_selection"
                )
            )

        external_pipeline = get_full_external_pipeline_or_raise(graphene_info, selector)

        if not external_pipeline.has_preset(preset_name):
            raise UserFacingGraphQLError(
                GraphenePresetNotFoundError(preset=preset_name, selector=selector)
            )

        preset = external_pipeline.get_preset(preset_name)

        return ExecutionParams(
            selector=selector.with_solid_selection(preset.solid_selection),
            run_config=preset.run_config,
            mode=preset.mode,
            execution_metadata=create_execution_metadata(
                graphql_execution_params.get("executionMetadata")
            ),
            step_keys=graphql_execution_params.get("stepKeys"),
        )

    return execution_params_from_graphql(graphql_execution_params)


def execution_params_from_graphql(graphql_execution_params):
    return ExecutionParams(
        selector=pipeline_selector_from_graphql(graphql_execution_params.get("selector")),
        run_config=parse_run_config_input(graphql_execution_params.get("runConfigData") or {}),
        mode=graphql_execution_params.get("mode"),
        execution_metadata=create_execution_metadata(
            graphql_execution_params.get("executionMetadata")
        ),
        step_keys=graphql_execution_params.get("stepKeys"),
    )


def create_execution_metadata(graphql_execution_metadata):
    return (
        ExecutionMetadata(
            run_id=graphql_execution_metadata.get("runId"),
            tags={t["key"]: t["value"] for t in graphql_execution_metadata.get("tags", [])},
            root_run_id=graphql_execution_metadata.get("rootRunId"),
            parent_run_id=graphql_execution_metadata.get("parentRunId"),
        )
        if graphql_execution_metadata
        else ExecutionMetadata(run_id=None, tags={})
    )


class GrapheneDeletePipelineRunSuccess(graphene.ObjectType):
    runId = graphene.NonNull(graphene.String)

    class Meta:
        name = "DeletePipelineRunSuccess"


class GrapheneDeletePipelineRunResult(graphene.Union):
    class Meta:
        types = (
            GrapheneDeletePipelineRunSuccess,
            GrapheneUnauthorizedError,
            GraphenePythonError,
            GrapheneRunNotFoundError,
        )
        name = "DeletePipelineRunResult"


class GrapheneDeleteRunMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneDeletePipelineRunResult)

    class Arguments:
        runId = graphene.NonNull(graphene.String)

    class Meta:
        name = "DeleteRunMutation"

    @capture_error
    @check_permission(Permissions.DELETE_PIPELINE_RUN)
    def mutate(self, graphene_info, **kwargs):
        run_id = kwargs["runId"]
        return delete_pipeline_run(graphene_info, run_id)


class GrapheneTerminatePipelineExecutionSuccess(graphene.Interface):
    run = graphene.Field(graphene.NonNull(GrapheneRun))

    class Meta:
        name = "TerminatePipelineExecutionSuccess"


class GrapheneTerminateRunSuccess(graphene.ObjectType):
    run = graphene.Field(graphene.NonNull(GrapheneRun))

    class Meta:
        interfaces = (GrapheneTerminatePipelineExecutionSuccess,)
        name = "TerminateRunSuccess"


class GrapheneTerminatePipelineExecutionFailure(graphene.Interface):
    run = graphene.NonNull(GrapheneRun)
    message = graphene.NonNull(graphene.String)

    class Meta:
        name = "TerminatePipelineExecutionFailure"


class GrapheneTerminateRunFailure(graphene.ObjectType):
    run = graphene.NonNull(GrapheneRun)
    message = graphene.NonNull(graphene.String)

    class Meta:
        interfaces = (GrapheneTerminatePipelineExecutionFailure,)
        name = "TerminateRunFailure"


class GrapheneTerminateRunResult(graphene.Union):
    class Meta:
        types = (
            GrapheneTerminateRunSuccess,
            GrapheneTerminateRunFailure,
            GrapheneRunNotFoundError,
            GrapheneUnauthorizedError,
            GraphenePythonError,
        )
        name = "TerminateRunResult"


@capture_error
def create_execution_params_and_launch_pipeline_exec(graphene_info, execution_params_dict):
    # refactored into a helper function here in order to wrap with @capture_error,
    # because create_execution_params may raise
    return launch_pipeline_execution(
        graphene_info,
        execution_params=create_execution_params(graphene_info, execution_params_dict),
    )


class GrapheneLaunchRunMutation(graphene.Mutation):

    Output = graphene.NonNull(GrapheneLaunchRunResult)

    class Arguments:
        executionParams = graphene.NonNull(GrapheneExecutionParams)

    class Meta:
        description = "Launch a run via the run launcher configured on the instance."
        name = "LaunchRunMutation"

    @capture_error
    @check_permission(Permissions.LAUNCH_PIPELINE_EXECUTION)
    def mutate(self, graphene_info, **kwargs):
        return create_execution_params_and_launch_pipeline_exec(
            graphene_info, kwargs["executionParams"]
        )


class GrapheneLaunchBackfillMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneLaunchBackfillResult)

    class Arguments:
        backfillParams = graphene.NonNull(GrapheneLaunchBackfillParams)

    class Meta:
        description = "Launches a set of partition backfill runs via the run launcher configured on the instance."
        name = "LaunchBackfillMutation"

    @capture_error
    @check_permission(Permissions.LAUNCH_PARTITION_BACKFILL)
    def mutate(self, graphene_info, **kwargs):
        return create_and_launch_partition_backfill(graphene_info, kwargs["backfillParams"])


class GrapheneCancelBackfillMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneCancelBackfillResult)

    class Arguments:
        backfillId = graphene.NonNull(graphene.String)

    class Meta:
        description = "Marks a partition backfill as canceled."
        name = "CancelBackfillMutation"

    @capture_error
    @check_permission(Permissions.CANCEL_PARTITION_BACKFILL)
    def mutate(self, graphene_info, **kwargs):
        return cancel_partition_backfill(graphene_info, kwargs["backfillId"])


class GrapheneResumeBackfillMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneResumeBackfillResult)

    class Arguments:
        backfillId = graphene.NonNull(graphene.String)

    class Meta:
        description = "Retries a set of partition backfill runs via the run launcher configured on the instance."
        name = "ResumeBackfillMutation"

    @capture_error
    @check_permission(Permissions.LAUNCH_PARTITION_BACKFILL)
    def mutate(self, graphene_info, **kwargs):
        return resume_partition_backfill(graphene_info, kwargs["backfillId"])


@capture_error
def create_execution_params_and_launch_pipeline_reexec(graphene_info, execution_params_dict):
    # refactored into a helper function here in order to wrap with @capture_error,
    # because create_execution_params may raise
    return launch_pipeline_reexecution(
        graphene_info,
        execution_params=create_execution_params(graphene_info, execution_params_dict),
    )


class GrapheneLaunchRunReexecutionMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneLaunchRunReexecutionResult)

    class Arguments:
        executionParams = graphene.NonNull(GrapheneExecutionParams)

    class Meta:
        description = "Re-launch a run via the run launcher configured on the instance"
        name = "LaunchRunReexecutionMutation"

    @capture_error
    @check_permission(Permissions.LAUNCH_PIPELINE_REEXECUTION)
    def mutate(self, graphene_info, **kwargs):
        return create_execution_params_and_launch_pipeline_reexec(
            graphene_info,
            execution_params_dict=kwargs["executionParams"],
        )


class GrapheneTerminateRunPolicy(graphene.Enum):
    # Default behavior: Only mark as canceled if the termination is successful, and after all
    # resources peforming the execution have been shut down.
    SAFE_TERMINATE = "SAFE_TERMINATE"

    # Immediately mark the pipelie as canceled, whether or not the termination was successful.
    # No guarantee that the execution has actually stopped.
    MARK_AS_CANCELED_IMMEDIATELY = "MARK_AS_CANCELED_IMMEDIATELY"

    class Meta:
        name = "TerminateRunPolicy"


class GrapheneTerminateRunMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneTerminateRunResult)

    class Arguments:
        runId = graphene.NonNull(graphene.String)
        terminatePolicy = graphene.Argument(GrapheneTerminateRunPolicy)

    class Meta:
        name = "TerminateRunMutation"

    @capture_error
    @check_permission(Permissions.TERMINATE_PIPELINE_EXECUTION)
    def mutate(self, graphene_info, **kwargs):
        return terminate_pipeline_execution(
            graphene_info.context.instance,
            kwargs["runId"],
            kwargs.get("terminatePolicy", GrapheneTerminateRunPolicy.SAFE_TERMINATE),
        )


class GrapheneReloadRepositoryLocationMutationResult(graphene.Union):
    class Meta:
        types = (
            GrapheneWorkspaceLocationEntry,
            GrapheneReloadNotSupported,
            GrapheneRepositoryLocationNotFound,
            GrapheneUnauthorizedError,
            GraphenePythonError,
        )
        name = "ReloadRepositoryLocationMutationResult"


class GrapheneShutdownRepositoryLocationSuccess(graphene.ObjectType):
    repositoryLocationName = graphene.NonNull(graphene.String)

    class Meta:
        name = "ShutdownRepositoryLocationSuccess"


class GrapheneShutdownRepositoryLocationMutationResult(graphene.Union):
    class Meta:
        types = (
            GrapheneShutdownRepositoryLocationSuccess,
            GrapheneRepositoryLocationNotFound,
            GrapheneUnauthorizedError,
            GraphenePythonError,
        )
        name = "ShutdownRepositoryLocationMutationResult"


class GrapheneReloadRepositoryLocationMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneReloadRepositoryLocationMutationResult)

    class Arguments:
        repositoryLocationName = graphene.NonNull(graphene.String)

    class Meta:
        name = "ReloadRepositoryLocationMutation"

    @capture_error
    @check_permission(Permissions.RELOAD_REPOSITORY_LOCATION)
    def mutate(self, graphene_info, **kwargs):
        location_name = kwargs["repositoryLocationName"]

        if not graphene_info.context.has_repository_location_name(location_name):
            return GrapheneRepositoryLocationNotFound(location_name)

        if not graphene_info.context.is_reload_supported(location_name):
            return GrapheneReloadNotSupported(location_name)

        # The current workspace context is a WorkspaceRequestContext, which contains a reference to the
        # repository locations that were present in the root IWorkspaceProcessContext the start of the
        # request. Reloading a repository location modifies the IWorkspaceProcessContext, rendeirng
        # our current WorkspaceRequestContext outdated. Therefore, `reload_repository_location` returns
        # an updated WorkspaceRequestContext for us to use.
        new_context = graphene_info.context.reload_repository_location(location_name)
        return GrapheneWorkspaceLocationEntry(new_context.get_location_entry(location_name))


class GrapheneShutdownRepositoryLocationMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneShutdownRepositoryLocationMutationResult)

    class Arguments:
        repositoryLocationName = graphene.NonNull(graphene.String)

    class Meta:
        name = "ShutdownRepositoryLocationMutation"

    @capture_error
    @check_permission(Permissions.RELOAD_REPOSITORY_LOCATION)
    def mutate(self, graphene_info, **kwargs):
        location_name = kwargs["repositoryLocationName"]

        if not graphene_info.context.has_repository_location_name(location_name):
            return GrapheneRepositoryLocationNotFound(location_name)

        if not graphene_info.context.is_shutdown_supported(location_name):
            raise Exception(f"Location {location_name} does not support shutting down via GraphQL")

        graphene_info.context.shutdown_repository_location(location_name)
        return GrapheneShutdownRepositoryLocationSuccess(repositoryLocationName=location_name)


class GrapheneReloadWorkspaceMutationResult(graphene.Union):
    class Meta:
        types = (
            GrapheneWorkspace,
            GrapheneUnauthorizedError,
            GraphenePythonError,
        )
        name = "ReloadWorkspaceMutationResult"


class GrapheneReloadWorkspaceMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneReloadWorkspaceMutationResult)

    class Meta:
        name = "ReloadWorkspaceMutation"

    @capture_error
    @check_permission(Permissions.RELOAD_WORKSPACE)
    def mutate(self, graphene_info, **_kwargs):
        new_context = graphene_info.context.reload_workspace()
        return fetch_workspace(new_context)


class GrapheneAssetWipeSuccess(graphene.ObjectType):
    assetKeys = non_null_list(GrapheneAssetKey)

    class Meta:
        name = "AssetWipeSuccess"


class GrapheneAssetWipeMutationResult(graphene.Union):
    class Meta:
        types = (
            GrapheneAssetNotFoundError,
            GrapheneUnauthorizedError,
            GraphenePythonError,
            GrapheneAssetWipeSuccess,
        )
        name = "AssetWipeMutationResult"


class GrapheneAssetWipeMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneAssetWipeMutationResult)

    class Arguments:
        assetKeys = graphene.Argument(non_null_list(GrapheneAssetKeyInput))

    class Meta:
        name = "AssetWipeMutation"

    @capture_error
    @check_permission(Permissions.WIPE_ASSETS)
    def mutate(self, graphene_info, **kwargs):
        return wipe_assets(
            graphene_info,
            [AssetKey.from_graphql_input(asset_key) for asset_key in kwargs["assetKeys"]],
        )


class GrapheneLogTelemetrySuccess(graphene.ObjectType):
    action = graphene.NonNull(graphene.String)

    class Meta:
        name = "LogTelemetrySuccess"


class GrapheneLogTelemetryMutationResult(graphene.Union):
    class Meta:
        types = (
            GrapheneLogTelemetrySuccess,
            GraphenePythonError,
        )
        name = "LogTelemetryMutationResult"


class GrapheneLogTelemetryMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneLogTelemetryMutationResult)

    class Arguments:
        action = graphene.Argument(graphene.NonNull(graphene.String))
        clientTime = graphene.Argument(graphene.NonNull(graphene.String))
        metadata = graphene.Argument(graphene.NonNull(graphene.String))

    class Meta:
        name = "LogTelemetryMutation"

    @capture_error
    def mutate(self, graphene_info, **kwargs):
        action = log_dagit_telemetry_event(
            graphene_info,
            action=kwargs["action"],
            client_time=kwargs["clientTime"],
            metadata=kwargs["metadata"],
        )
        return action


class GrapheneDagitMutation(graphene.ObjectType):
    class Meta:
        name = "DagitMutation"

    launch_pipeline_execution = GrapheneLaunchRunMutation.Field()
    launch_run = GrapheneLaunchRunMutation.Field()
    launch_pipeline_reexecution = GrapheneLaunchRunReexecutionMutation.Field()
    launch_run_reexecution = GrapheneLaunchRunReexecutionMutation.Field()
    start_schedule = GrapheneStartScheduleMutation.Field()
    stop_running_schedule = GrapheneStopRunningScheduleMutation.Field()
    start_sensor = GrapheneStartSensorMutation.Field()
    set_sensor_cursor = GrapheneSetSensorCursorMutation.Field()
    stop_sensor = GrapheneStopSensorMutation.Field()
    terminate_pipeline_execution = GrapheneTerminateRunMutation.Field()
    terminate_run = GrapheneTerminateRunMutation.Field()
    delete_pipeline_run = GrapheneDeleteRunMutation.Field()
    delete_run = GrapheneDeleteRunMutation.Field()
    reload_repository_location = GrapheneReloadRepositoryLocationMutation.Field()
    reload_workspace = GrapheneReloadWorkspaceMutation.Field()
    shutdown_repository_location = GrapheneShutdownRepositoryLocationMutation.Field()
    wipe_assets = GrapheneAssetWipeMutation.Field()
    launch_partition_backfill = GrapheneLaunchBackfillMutation.Field()
    resume_partition_backfill = GrapheneResumeBackfillMutation.Field()
    cancel_partition_backfill = GrapheneCancelBackfillMutation.Field()
    log_telemetry = GrapheneLogTelemetryMutation.Field()



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['TestAssetAwareEventLog.test_all_asset_keys[asset_aware_instance_in_process_env] 1'] = {
    'assetsOrError': {
        '__typename': 'AssetConnection',
        'nodes': [
            {
                'key': {
                    'path': [
                        'a'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_1'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_2'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_3'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_one'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_two'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_yields_observation'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'b'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'c'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'downstream_static_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'downstream_time_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'dummy_source_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'first_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'hanging_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'never_runs_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'upstream_static_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'upstream_time_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'yield_partition_materialization'
                    ]
                }
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_all_asset_keys[postgres_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetsOrError': {
        '__typename': 'AssetConnection',
        'nodes': [
            {
                'key': {
                    'path': [
                        'a'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_1'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_2'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_3'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_one'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_two'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_yields_observation'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'b'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'c'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'downstream_static_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'downstream_time_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'dummy_source_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'first_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'hanging_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'never_runs_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'upstream_static_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'upstream_time_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'yield_partition_materialization'
                    ]
                }
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_all_asset_keys[sqlite_with_default_run_launcher_deployed_grpc_env] 1'] = {
    'assetsOrError': {
        '__typename': 'AssetConnection',
        'nodes': [
            {
                'key': {
                    'path': [
                        'a'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_1'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_2'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_3'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_one'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_two'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_yields_observation'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'b'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'c'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'downstream_static_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'downstream_time_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'dummy_source_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'first_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'hanging_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'never_runs_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'upstream_static_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'upstream_time_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'yield_partition_materialization'
                    ]
                }
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_all_asset_keys[sqlite_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetsOrError': {
        '__typename': 'AssetConnection',
        'nodes': [
            {
                'key': {
                    'path': [
                        'a'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_1'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_2'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_3'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_one'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_two'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'asset_yields_observation'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'b'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'c'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'downstream_static_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'downstream_time_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'dummy_source_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'first_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'hanging_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'never_runs_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'upstream_static_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'upstream_time_partitioned_asset'
                    ]
                }
            },
            {
                'key': {
                    'path': [
                        'yield_partition_materialization'
                    ]
                }
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_asset_op[asset_aware_instance_in_process_env] 1'] = {
    'assetOrError': {
        'definition': {
            'op': {
                'description': None,
                'inputDefinitions': [
                    {
                        'name': 'asset_one'
                    }
                ],
                'name': 'asset_two',
                'outputDefinitions': [
                    {
                        'name': 'result'
                    }
                ]
            }
        }
    }
}

snapshots['TestAssetAwareEventLog.test_asset_op[sqlite_with_default_run_launcher_deployed_grpc_env] 1'] = {
    'assetOrError': {
        'definition': {
            'op': {
                'description': None,
                'inputDefinitions': [
                    {
                        'name': 'asset_one'
                    }
                ],
                'name': 'asset_two',
                'outputDefinitions': [
                    {
                        'name': 'result'
                    }
                ]
            }
        }
    }
}

snapshots['TestAssetAwareEventLog.test_asset_op[sqlite_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetOrError': {
        'definition': {
            'op': {
                'description': None,
                'inputDefinitions': [
                    {
                        'name': 'asset_one'
                    }
                ],
                'name': 'asset_two',
                'outputDefinitions': [
                    {
                        'name': 'result'
                    }
                ]
            }
        }
    }
}

snapshots['TestAssetAwareEventLog.test_get_asset_key_lineage[asset_aware_instance_in_process_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'assetLineage': [
                ],
                'label': 'b'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_asset_key_lineage[postgres_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'assetLineage': [
                ],
                'label': 'b'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_asset_key_lineage[sqlite_with_default_run_launcher_deployed_grpc_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'assetLineage': [
                ],
                'label': 'b'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_asset_key_lineage[sqlite_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'assetLineage': [
                ],
                'label': 'b'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_asset_key_materialization[asset_aware_instance_in_process_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'assetLineage': [
                ],
                'label': 'a'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_asset_key_materialization[postgres_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'assetLineage': [
                ],
                'label': 'a'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_asset_key_materialization[sqlite_with_default_run_launcher_deployed_grpc_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'assetLineage': [
                ],
                'label': 'a'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_asset_key_materialization[sqlite_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'assetLineage': [
                ],
                'label': 'a'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_asset_key_not_found[asset_aware_instance_in_process_env] 1'] = {
    'assetOrError': {
        '__typename': 'AssetNotFoundError'
    }
}

snapshots['TestAssetAwareEventLog.test_get_asset_key_not_found[postgres_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetOrError': {
        '__typename': 'AssetNotFoundError'
    }
}

snapshots['TestAssetAwareEventLog.test_get_asset_key_not_found[sqlite_with_default_run_launcher_deployed_grpc_env] 1'] = {
    'assetOrError': {
        '__typename': 'AssetNotFoundError'
    }
}

snapshots['TestAssetAwareEventLog.test_get_asset_key_not_found[sqlite_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetOrError': {
        '__typename': 'AssetNotFoundError'
    }
}

snapshots['TestAssetAwareEventLog.test_get_partitioned_asset_key_lineage[asset_aware_instance_in_process_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'assetLineage': [
                ],
                'label': 'b'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_partitioned_asset_key_lineage[postgres_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'assetLineage': [
                ],
                'label': 'b'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_partitioned_asset_key_lineage[sqlite_with_default_run_launcher_deployed_grpc_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'assetLineage': [
                ],
                'label': 'b'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_partitioned_asset_key_lineage[sqlite_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'assetLineage': [
                ],
                'label': 'b'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_partitioned_asset_key_materialization[asset_aware_instance_in_process_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'label': 'a',
                'partition': 'partition_1'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_partitioned_asset_key_materialization[postgres_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'label': 'a',
                'partition': 'partition_1'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_partitioned_asset_key_materialization[sqlite_with_default_run_launcher_deployed_grpc_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'label': 'a',
                'partition': 'partition_1'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_get_partitioned_asset_key_materialization[sqlite_with_default_run_launcher_managed_grpc_env] 1'] = {
    'assetOrError': {
        'assetMaterializations': [
            {
                'label': 'a',
                'partition': 'partition_1'
            }
        ]
    }
}

snapshots['TestAssetAwareEventLog.test_op_assets[asset_aware_instance_in_process_env] 1'] = {
    'repositoryOrError': {
        'usedSolid': {
            'definition': {
                'assetNodes': [
                    {
                        'assetKey': {
                            'path': [
                                'asset_two'
                            ]
                        }
                    }
                ]
            }
        }
    }
}

snapshots['TestAssetAwareEventLog.test_op_assets[sqlite_with_default_run_launcher_deployed_grpc_env] 1'] = {
    'repositoryOrError': {
        'usedSolid': {
            'definition': {
                'assetNodes': [
                    {
                        'assetKey': {
                            'path': [
                                'asset_two'
                            ]
                        }
                    }
                ]
            }
        }
    }
}

snapshots['TestAssetAwareEventLog.test_op_assets[sqlite_with_default_run_launcher_managed_grpc_env] 1'] = {
    'repositoryOrError': {
        'usedSolid': {
            'definition': {
                'assetNodes': [
                    {
                        'assetKey': {
                            'path': [
                                'asset_two'
                            ]
                        }
                    }
                ]
            }
        }
    }
}



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['TestRetryExecution.test_pipeline_reexecution_info_query[postgres_with_default_run_launcher_deployed_grpc_env] 1'] = [
    'sum_sq_solid'
]

snapshots['TestRetryExecution.test_pipeline_reexecution_info_query[postgres_with_default_run_launcher_managed_grpc_env] 1'] = [
    'sum_sq_solid'
]

snapshots['TestRetryExecution.test_pipeline_reexecution_info_query[sqlite_with_default_run_launcher_deployed_grpc_env] 1'] = [
    'sum_sq_solid'
]

snapshots['TestRetryExecution.test_pipeline_reexecution_info_query[sqlite_with_default_run_launcher_managed_grpc_env] 1'] = [
    'sum_sq_solid'
]



import graphene

from dagster import check
from dagster.core.snap import ConfigSchemaSnapshot, ResourceDefSnap

from ..config_types import GrapheneConfigTypeField


class GrapheneResource(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    configField = graphene.Field(GrapheneConfigTypeField)

    class Meta:
        name = "Resource"

    def __init__(self, config_schema_snapshot, resource_def_snap):
        super().__init__()
        self._config_schema_snapshot = check.inst_param(
            config_schema_snapshot, "config_schema_snapshot", ConfigSchemaSnapshot
        )
        self._resource_def_snap = check.inst_param(
            resource_def_snap, "resource_def_snap", ResourceDefSnap
        )
        self.name = resource_def_snap.name
        self.description = resource_def_snap.description

    def resolve_configField(self, _graphene_info):
        if (
            self._resource_def_snap.config_field_snap
            # config type may not be present if mode config mapped
            and self._config_schema_snapshot.has_config_snap(
                self._resource_def_snap.config_field_snap.type_key
            )
        ):
            return GrapheneConfigTypeField(
                config_schema_snapshot=self._config_schema_snapshot,
                field_snap=self._resource_def_snap.config_field_snap,
            )

        return None



import tempfile
import time
from collections import namedtuple
from contextlib import contextmanager
from unittest.mock import Mock

import pytest
from dagster_graphql.implementation.pipeline_run_storage import PipelineRunObservableSubscribe
from dagster_tests.core_tests.storage_tests.test_polling_event_watcher import (
    SqlitePollingEventLogStorage,
)

from dagster.core.events import DagsterEvent, DagsterEventType, EngineEventData
from dagster.core.events.log import EventLogEntry
from dagster.core.storage.event_log import SqlEventLogStorage
from dagster.core.test_utils import instance_for_test


@contextmanager
def create_test_instance_and_storage():
    with tempfile.TemporaryDirectory() as tmpdir_path:
        with instance_for_test(
            temp_dir=tmpdir_path,
            overrides={
                "event_log_storage": {
                    "module": "dagster_tests.core_tests.storage_tests.test_polling_event_watcher",
                    "class": "SqlitePollingEventLogStorage",
                    "config": {"base_dir": tmpdir_path},
                }
            },
        ) as instance:
            yield (instance, instance._event_storage)  # pylint: disable=protected-access


RUN_ID = "foo"


class EventStorer:
    def __init__(self, storage: SqlEventLogStorage):
        self._storage = storage
        self._counter = 0

    def store_n_events(self, n: int):
        for _ in range(n):
            self._counter += 1
            self._storage.store_event(self.create_event(self._counter))

    @staticmethod
    def create_event(count: int, run_id: str = RUN_ID):
        return EventLogEntry(
            error_info=None,
            user_message=str(count),
            level="debug",
            run_id=run_id,
            timestamp=time.time(),
            dagster_event=DagsterEvent(
                DagsterEventType.ENGINE_EVENT.value,
                "nonce",
                event_specific_data=EngineEventData.in_process(999),
            ),
        )


NumEventsAndCursor = namedtuple("NumEventsAndCursor", ["num_events_before_watch", "after_cursor"])

MAX_NUM_EVENTS_BEFORE_WATCH = 2
MAX_NUM_EVENTS_AFTER_WATCH = 2


@pytest.mark.parametrize(
    "before_watch_config",
    [
        NumEventsAndCursor(num_events_before_watch, after_cursor)
        for num_events_before_watch in range(0, MAX_NUM_EVENTS_BEFORE_WATCH + 1)
        for after_cursor in range(-1, num_events_before_watch + 1)
    ],
)
@pytest.mark.parametrize("num_events_after_watch", list(range(1, MAX_NUM_EVENTS_AFTER_WATCH + 1)))
def test_using_instance(before_watch_config: NumEventsAndCursor, num_events_after_watch: int):
    total_num_events: int = before_watch_config.num_events_before_watch + num_events_after_watch
    with create_test_instance_and_storage() as (instance, storage):
        # set up instance & write `before_watch_config.num_events_before_watch` to event_log
        assert isinstance(storage, SqlitePollingEventLogStorage)
        observable_subscribe = PipelineRunObservableSubscribe(
            instance, RUN_ID, after_cursor=before_watch_config.after_cursor
        )
        event_storer = EventStorer(storage)
        event_storer.store_n_events(before_watch_config.num_events_before_watch)

        # start watching events & write `num_events_after_watch` to event_log
        observable_subscribe(Mock())
        event_storer.store_n_events(num_events_after_watch)
        call_args = observable_subscribe.observer.on_next.call_args_list

        # wait until all events have been captured
        most_recent_event_processed = lambda: int(call_args[-1][0][0][0][-1].user_message)
        attempts = 10
        while (
            len(call_args) == 0 or most_recent_event_processed() < total_num_events
        ) and attempts > 0:
            time.sleep(0.1)
            attempts -= 1

        # ensure all expected events captured, no duplicates, etc.
        events_list = [
            [event_record.user_message for event_record in call[0][0][0]] for call in call_args
        ]
        flattened_events_list = [int(message) for lst in events_list for message in lst]
        # PipelineRunObservableSubscribe requests ids > after_cursor + 1
        beginning_cursor = before_watch_config.after_cursor + 2
        assert flattened_events_list == list(
            range(
                beginning_cursor,
                total_num_events + 1,
            )
        )



from dagster_graphql.client.query import (
    LAUNCH_PIPELINE_EXECUTION_MUTATION,
    LAUNCH_PIPELINE_REEXECUTION_MUTATION,
)
from dagster_graphql.test.utils import (
    execute_dagster_graphql,
    execute_dagster_graphql_and_finish_runs,
    infer_pipeline_selector,
)

from dagster.core.storage.tags import RESUME_RETRY_TAG

from .utils import (
    get_all_logs_for_finished_run_via_subscription,
    step_did_fail,
    step_did_not_run,
    step_did_succeed,
)


def test_dynamic_resume_reexecution(graphql_context):
    selector = infer_pipeline_selector(graphql_context, "dynamic_pipeline")
    result = execute_dagster_graphql_and_finish_runs(
        graphql_context,
        LAUNCH_PIPELINE_EXECUTION_MUTATION,
        variables={
            "executionParams": {
                "selector": selector,
                "runConfigData": {
                    "solids": {"multiply_inputs": {"inputs": {"should_fail": {"value": True}}}},
                },
                "mode": "default",
            }
        },
    )

    assert not result.errors
    assert result.data
    assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
    assert result.data["launchPipelineExecution"]["run"]["pipeline"]["name"] == "dynamic_pipeline"

    parent_run_id = result.data["launchPipelineExecution"]["run"]["runId"]

    logs = get_all_logs_for_finished_run_via_subscription(graphql_context, parent_run_id)[
        "pipelineRunLogs"
    ]["messages"]

    assert step_did_succeed(logs, "emit")
    assert step_did_succeed(logs, "multiply_inputs[0]")
    assert step_did_succeed(logs, "multiply_inputs[1]")
    assert step_did_fail(logs, "multiply_inputs[2]")
    assert step_did_succeed(logs, "multiply_by_two[0]")
    assert step_did_succeed(logs, "multiply_by_two[1]")

    retry_one = execute_dagster_graphql_and_finish_runs(
        graphql_context,
        LAUNCH_PIPELINE_REEXECUTION_MUTATION,
        variables={
            "executionParams": {
                "mode": "default",
                "selector": selector,
                "runConfigData": {
                    "solids": {"multiply_inputs": {"inputs": {"should_fail": {"value": True}}}},
                    "execution": {"multiprocess": {}},
                },
                "executionMetadata": {
                    "rootRunId": parent_run_id,
                    "parentRunId": parent_run_id,
                    "tags": [{"key": RESUME_RETRY_TAG, "value": "true"}],
                },
            }
        },
    )
    assert not retry_one.errors
    assert retry_one.data
    assert (
        retry_one.data["launchPipelineReexecution"]["__typename"] == "LaunchRunSuccess"
    ), retry_one.data["launchPipelineReexecution"].get("message")

    run_id = retry_one.data["launchPipelineReexecution"]["run"]["runId"]

    logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
        "pipelineRunLogs"
    ]["messages"]

    assert step_did_not_run(logs, "emit")
    assert step_did_not_run(logs, "multiply_inputs[0]")
    assert step_did_not_run(logs, "multiply_inputs[1]")
    assert step_did_succeed(logs, "multiply_inputs[2]")

    assert step_did_not_run(logs, "multiply_by_two[0]")
    assert step_did_not_run(logs, "multiply_by_two[1]")
    assert step_did_succeed(logs, "multiply_by_two[2]")
    assert step_did_succeed(logs, "double_total")


def test_dynamic_full_reexecution(graphql_context):
    selector = infer_pipeline_selector(graphql_context, "dynamic_pipeline")
    result = execute_dagster_graphql_and_finish_runs(
        graphql_context,
        LAUNCH_PIPELINE_EXECUTION_MUTATION,
        variables={
            "executionParams": {
                "selector": selector,
                "runConfigData": {
                    "solids": {"multiply_inputs": {"inputs": {"should_fail": {"value": True}}}},
                },
                "mode": "default",
            }
        },
    )

    assert not result.errors
    assert result.data
    assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
    assert result.data["launchPipelineExecution"]["run"]["pipeline"]["name"] == "dynamic_pipeline"

    parent_run_id = result.data["launchPipelineExecution"]["run"]["runId"]

    logs = get_all_logs_for_finished_run_via_subscription(graphql_context, parent_run_id)[
        "pipelineRunLogs"
    ]["messages"]

    assert step_did_succeed(logs, "emit")
    assert step_did_succeed(logs, "multiply_inputs[0]")
    assert step_did_succeed(logs, "multiply_inputs[1]")
    assert step_did_fail(logs, "multiply_inputs[2]")
    assert step_did_succeed(logs, "multiply_by_two[0]")
    assert step_did_succeed(logs, "multiply_by_two[1]")

    retry_one = execute_dagster_graphql_and_finish_runs(
        graphql_context,
        LAUNCH_PIPELINE_REEXECUTION_MUTATION,
        variables={
            "executionParams": {
                "mode": "default",
                "selector": selector,
                "runConfigData": {
                    "solids": {"multiply_inputs": {"inputs": {"should_fail": {"value": True}}}},
                    "execution": {"multiprocess": {}},
                },
                "executionMetadata": {
                    "rootRunId": parent_run_id,
                    "parentRunId": parent_run_id,
                },
                "stepKeys": None,
            }
        },
    )
    assert not retry_one.errors
    assert retry_one.data
    assert (
        retry_one.data["launchPipelineReexecution"]["__typename"] == "LaunchRunSuccess"
    ), retry_one.data["launchPipelineReexecution"].get("message")

    run_id = retry_one.data["launchPipelineReexecution"]["run"]["runId"]

    logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
        "pipelineRunLogs"
    ]["messages"]

    assert step_did_succeed(logs, "emit")
    assert step_did_succeed(logs, "multiply_inputs[0]")
    assert step_did_succeed(logs, "multiply_inputs[1]")
    assert step_did_succeed(logs, "multiply_inputs[2]")

    assert step_did_succeed(logs, "multiply_by_two[0]")
    assert step_did_succeed(logs, "multiply_by_two[1]")
    assert step_did_succeed(logs, "multiply_by_two[2]")
    assert step_did_succeed(logs, "double_total")


def test_dynamic_subset(graphql_context):
    selector = infer_pipeline_selector(graphql_context, "dynamic_pipeline")
    result = execute_dagster_graphql_and_finish_runs(
        graphql_context,
        LAUNCH_PIPELINE_EXECUTION_MUTATION,
        variables={
            "executionParams": {
                "selector": selector,
                "runConfigData": {
                    "solids": {"multiply_inputs": {"inputs": {"should_fail": {"value": True}}}},
                },
                "mode": "default",
            }
        },
    )

    assert not result.errors
    assert result.data
    assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
    assert result.data["launchPipelineExecution"]["run"]["pipeline"]["name"] == "dynamic_pipeline"

    parent_run_id = result.data["launchPipelineExecution"]["run"]["runId"]

    logs = get_all_logs_for_finished_run_via_subscription(graphql_context, parent_run_id)[
        "pipelineRunLogs"
    ]["messages"]

    assert step_did_succeed(logs, "emit")
    assert step_did_succeed(logs, "multiply_inputs[0]")
    assert step_did_succeed(logs, "multiply_inputs[1]")
    assert step_did_fail(logs, "multiply_inputs[2]")
    assert step_did_succeed(logs, "multiply_by_two[0]")
    assert step_did_succeed(logs, "multiply_by_two[1]")

    retry_one = execute_dagster_graphql_and_finish_runs(
        graphql_context,
        LAUNCH_PIPELINE_REEXECUTION_MUTATION,
        variables={
            "executionParams": {
                "mode": "default",
                "selector": selector,
                "runConfigData": {
                    "solids": {"multiply_inputs": {"inputs": {"should_fail": {"value": True}}}},
                    "execution": {"multiprocess": {}},
                },
                "executionMetadata": {
                    "rootRunId": parent_run_id,
                    "parentRunId": parent_run_id,
                },
                # manual version of from-failure above
                "stepKeys": [
                    "multiply_inputs[2]",
                    "multiply_by_two[2]",
                    "sum_numbers",
                    "double_total",
                ],
            }
        },
    )
    assert not retry_one.errors
    assert retry_one.data
    assert (
        retry_one.data["launchPipelineReexecution"]["__typename"] == "LaunchRunSuccess"
    ), retry_one.data["launchPipelineReexecution"].get("message")

    run_id = retry_one.data["launchPipelineReexecution"]["run"]["runId"]

    logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
        "pipelineRunLogs"
    ]["messages"]

    assert step_did_not_run(logs, "emit")
    assert step_did_not_run(logs, "multiply_inputs[0]")
    assert step_did_not_run(logs, "multiply_inputs[1]")
    assert step_did_succeed(logs, "multiply_inputs[2]")

    assert step_did_not_run(logs, "multiply_by_two[0]")
    assert step_did_not_run(logs, "multiply_by_two[1]")
    assert step_did_succeed(logs, "multiply_by_two[2]")
    assert step_did_succeed(logs, "double_total")


DEP_QUERY = """
query PresetsQuery($selector: PipelineSelector!) {
  pipelineOrError(params: $selector) {
    ... on Pipeline {
      name
      solids {
        name
        isDynamicMapped
        inputs {
          isDynamicCollect
          dependsOn {
            solid { name }
          }
        }
        outputs {
          definition {
            name
            isDynamic
          }
        }
      }
    }
  }
}
"""


def test_dynamic_dep_fields(graphql_context):
    selector = infer_pipeline_selector(graphql_context, "dynamic_pipeline")
    result = execute_dagster_graphql(graphql_context, DEP_QUERY, variables={"selector": selector})
    assert not result.errors
    solids = {solid["name"]: solid for solid in result.data["pipelineOrError"]["solids"]}

    assert solids["emit_ten"]["isDynamicMapped"] == False
    assert solids["emit_ten"]["outputs"][0]["definition"]["isDynamic"] == False

    assert solids["emit"]["isDynamicMapped"] == False
    assert solids["emit"]["outputs"][0]["definition"]["isDynamic"] == True

    assert solids["multiply_inputs"]["isDynamicMapped"] == True

    assert solids["multiply_by_two"]["isDynamicMapped"] == True

    assert solids["sum_numbers"]["isDynamicMapped"] == False
    assert solids["sum_numbers"]["inputs"][0]["isDynamicCollect"] == True

    assert solids["double_total"]["isDynamicMapped"] == False
    assert solids["double_total"]["inputs"][0]["isDynamicCollect"] == False



import logging

import graphene

from dagster import check


class GrapheneLogLevel(graphene.Enum):
    CRITICAL = "CRITICAL"
    ERROR = "ERROR"
    INFO = "INFO"
    WARNING = "WARNING"
    DEBUG = "DEBUG"

    class Meta:
        name = "LogLevel"

    @classmethod
    def from_level(cls, level):
        check.int_param(level, "level")
        if level == logging.CRITICAL:
            return GrapheneLogLevel.CRITICAL
        elif level == logging.ERROR:
            return GrapheneLogLevel.ERROR
        elif level == logging.INFO:
            return GrapheneLogLevel.INFO
        elif level == logging.WARNING:
            return GrapheneLogLevel.WARNING
        elif level == logging.DEBUG:
            return GrapheneLogLevel.DEBUG
        else:
            check.failed("Invalid log level: {level}".format(level=level))



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['TestGraphs.test_basic_graphs[non_launchable_in_memory_instance_lazy_repository] 1'] = {
    'graphOrError': {
        '__typename': 'Graph',
        'name': 'composed_graph',
        'solidHandles': [
            {
                'handleID': 'simple_graph',
                'solid': {
                    'name': 'simple_graph'
                }
            },
            {
                'handleID': 'simple_graph.noop_solid',
                'solid': {
                    'name': 'noop_solid'
                }
            }
        ]
    }
}

snapshots['TestGraphs.test_basic_graphs[non_launchable_in_memory_instance_managed_grpc_env] 1'] = {
    'graphOrError': {
        '__typename': 'Graph',
        'name': 'composed_graph',
        'solidHandles': [
            {
                'handleID': 'simple_graph',
                'solid': {
                    'name': 'simple_graph'
                }
            },
            {
                'handleID': 'simple_graph.noop_solid',
                'solid': {
                    'name': 'noop_solid'
                }
            }
        ]
    }
}

snapshots['TestGraphs.test_basic_graphs[non_launchable_in_memory_instance_multi_location] 1'] = {
    'graphOrError': {
        '__typename': 'Graph',
        'name': 'composed_graph',
        'solidHandles': [
            {
                'handleID': 'simple_graph',
                'solid': {
                    'name': 'simple_graph'
                }
            },
            {
                'handleID': 'simple_graph.noop_solid',
                'solid': {
                    'name': 'noop_solid'
                }
            }
        ]
    }
}

snapshots['TestGraphs.test_basic_graphs[non_launchable_postgres_instance_lazy_repository] 1'] = {
    'graphOrError': {
        '__typename': 'Graph',
        'name': 'composed_graph',
        'solidHandles': [
            {
                'handleID': 'simple_graph',
                'solid': {
                    'name': 'simple_graph'
                }
            },
            {
                'handleID': 'simple_graph.noop_solid',
                'solid': {
                    'name': 'noop_solid'
                }
            }
        ]
    }
}

snapshots['TestGraphs.test_basic_graphs[non_launchable_postgres_instance_managed_grpc_env] 1'] = {
    'graphOrError': {
        '__typename': 'Graph',
        'name': 'composed_graph',
        'solidHandles': [
            {
                'handleID': 'simple_graph',
                'solid': {
                    'name': 'simple_graph'
                }
            },
            {
                'handleID': 'simple_graph.noop_solid',
                'solid': {
                    'name': 'noop_solid'
                }
            }
        ]
    }
}

snapshots['TestGraphs.test_basic_graphs[non_launchable_postgres_instance_multi_location] 1'] = {
    'graphOrError': {
        '__typename': 'Graph',
        'name': 'composed_graph',
        'solidHandles': [
            {
                'handleID': 'simple_graph',
                'solid': {
                    'name': 'simple_graph'
                }
            },
            {
                'handleID': 'simple_graph.noop_solid',
                'solid': {
                    'name': 'noop_solid'
                }
            }
        ]
    }
}

snapshots['TestGraphs.test_basic_graphs[non_launchable_sqlite_instance_deployed_grpc_env] 1'] = {
    'graphOrError': {
        '__typename': 'Graph',
        'name': 'composed_graph',
        'solidHandles': [
            {
                'handleID': 'simple_graph',
                'solid': {
                    'name': 'simple_graph'
                }
            },
            {
                'handleID': 'simple_graph.noop_solid',
                'solid': {
                    'name': 'noop_solid'
                }
            }
        ]
    }
}

snapshots['TestGraphs.test_basic_graphs[non_launchable_sqlite_instance_lazy_repository] 1'] = {
    'graphOrError': {
        '__typename': 'Graph',
        'name': 'composed_graph',
        'solidHandles': [
            {
                'handleID': 'simple_graph',
                'solid': {
                    'name': 'simple_graph'
                }
            },
            {
                'handleID': 'simple_graph.noop_solid',
                'solid': {
                    'name': 'noop_solid'
                }
            }
        ]
    }
}

snapshots['TestGraphs.test_basic_graphs[non_launchable_sqlite_instance_managed_grpc_env] 1'] = {
    'graphOrError': {
        '__typename': 'Graph',
        'name': 'composed_graph',
        'solidHandles': [
            {
                'handleID': 'simple_graph',
                'solid': {
                    'name': 'simple_graph'
                }
            },
            {
                'handleID': 'simple_graph.noop_solid',
                'solid': {
                    'name': 'noop_solid'
                }
            }
        ]
    }
}

snapshots['TestGraphs.test_basic_graphs[non_launchable_sqlite_instance_multi_location] 1'] = {
    'graphOrError': {
        '__typename': 'Graph',
        'name': 'composed_graph',
        'solidHandles': [
            {
                'handleID': 'simple_graph',
                'solid': {
                    'name': 'simple_graph'
                }
            },
            {
                'handleID': 'simple_graph.noop_solid',
                'solid': {
                    'name': 'noop_solid'
                }
            }
        ]
    }
}



import os
from typing import Iterator

import pytest
from dagster_graphql.schema import create_schema
from gql import Client, gql

from dagster.utils import file_relative_path


@pytest.fixture(scope="session", name="validator_client")
def get_validator_client() -> Client:
    return Client(schema=create_schema())


def generate_legacy_query_strings() -> Iterator[str]:
    for (dir_name, _, file_names) in os.walk(file_relative_path(__file__, "./query_snapshots")):
        for file_name in file_names:
            if file_name.endswith(".graphql"):
                with open(os.path.join(dir_name, file_name)) as f:
                    yield f.read()


@pytest.mark.parametrize("query_str", generate_legacy_query_strings())
def test_backcompat(validator_client: Client, query_str: str):
    validator_client.validate(gql(query_str))



from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from .composites_query import (
    COMPOSITES_QUERY,
    COMPOSITES_QUERY_NESTED_DEPENDS_ON_DEPENDS_BY_CORE,
    NESTED_INPUT_DEPENDS_ON,
    NESTED_OUTPUT_DEPENDED_BY,
    PARENT_ID_QUERY,
    SOLID_ID_QUERY,
)
from .graphql_context_test_suite import NonLaunchableGraphQLContextTestMatrix

# 10 total solids in the composite pipeline:
#
# (+1) \
#       (+2)
# (+1) /    \
#            (+4)
# (+1) \    /
#       (+2)
# (+1) /
#
#       (/2)
#           \
#            (/4)
#           /
#       (/2)


# this only needs non_launchable variants since they never execute anything
class TestComposites(NonLaunchableGraphQLContextTestMatrix):
    def test_composites(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, "composites_pipeline")
        result = execute_dagster_graphql(graphql_context, COMPOSITES_QUERY, {"selector": selector})
        handle_map = {}
        for obj in result.data["pipelineOrError"]["solidHandles"]:
            handle_map[obj["handleID"]] = obj["solid"]

        assert len(handle_map) == 10

        snapshot.assert_match(result.data)

    def test_parent_id_arg(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, "composites_pipeline")
        result = execute_dagster_graphql(graphql_context, PARENT_ID_QUERY, {"selector": selector})
        assert len(result.data["pipelineOrError"]["solidHandles"]) == 10

        result = execute_dagster_graphql(
            graphql_context, PARENT_ID_QUERY, {"selector": selector, "parentHandleID": ""}
        )
        assert len(result.data["pipelineOrError"]["solidHandles"]) == 2

        result = execute_dagster_graphql(
            graphql_context, PARENT_ID_QUERY, {"selector": selector, "parentHandleID": "add_four"}
        )
        assert len(result.data["pipelineOrError"]["solidHandles"]) == 2

        result = execute_dagster_graphql(
            graphql_context,
            PARENT_ID_QUERY,
            {"selector": selector, "parentHandleID": "add_four.adder_1"},
        )
        assert len(result.data["pipelineOrError"]["solidHandles"]) == 2

        result = execute_dagster_graphql(
            graphql_context,
            PARENT_ID_QUERY,
            {"selector": selector, "parentHandleID": "add_four.doot"},
        )
        assert len(result.data["pipelineOrError"]["solidHandles"]) == 0

    def test_solid_id(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, "composites_pipeline")
        result = execute_dagster_graphql(
            graphql_context, SOLID_ID_QUERY, {"selector": selector, "id": "add_four"}
        )
        assert result.data["pipelineOrError"]["solidHandle"]["handleID"] == "add_four"

        result = execute_dagster_graphql(
            graphql_context,
            SOLID_ID_QUERY,
            {"selector": selector, "id": "add_four.adder_1.adder_1"},
        )
        assert (
            result.data["pipelineOrError"]["solidHandle"]["handleID"] == "add_four.adder_1.adder_1"
        )

        result = execute_dagster_graphql(
            graphql_context, SOLID_ID_QUERY, {"selector": selector, "id": "bonkahog"}
        )
        assert result.data["pipelineOrError"]["solidHandle"] == None

    def test_recurse_composites_depends(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, "composites_pipeline")
        execute_dagster_graphql(
            graphql_context,
            COMPOSITES_QUERY_NESTED_DEPENDS_ON_DEPENDS_BY_CORE + NESTED_INPUT_DEPENDS_ON,
            {"selector": selector},
        )

        execute_dagster_graphql(
            graphql_context,
            COMPOSITES_QUERY_NESTED_DEPENDS_ON_DEPENDS_BY_CORE + NESTED_OUTPUT_DEPENDED_BY,
            {"selector": selector},
        )

    def test_composed_graph(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, "composed_graph")
        result = execute_dagster_graphql(graphql_context, COMPOSITES_QUERY, {"selector": selector})
        handle_map = {}
        for obj in result.data["pipelineOrError"]["solidHandles"]:
            handle_map[obj["handleID"]] = obj["solid"]

        assert len(handle_map) == 2



from dagster_graphql.test.utils import execute_dagster_graphql, infer_repository_selector

from .graphql_context_test_suite import NonLaunchableGraphQLContextTestMatrix

REPOSITORY_QUERY = """
query {
   workspaceOrError {
      __typename
      ... on Workspace {
        locationEntries {
          __typename
          id
          name
          locationOrLoadError {
            __typename
            ... on RepositoryLocation {
                id
                name
                repositories {
                    name
                    pipelines {
                        name
                        graphName
                    }
                }
                isReloadSupported
            }
            ... on PythonError {
              message
            }
          }
        }
      }
      ... on PythonError {
          message
          stack
      }
    }
}
"""

GRAPH_QUERY = """
query GraphQuery($selector: GraphSelector!) {
  graphOrError(selector: $selector) {
    __typename
    ... on Graph {
      name
      solidHandles {
        handleID
        solid {
          name
        }
      }
    }
    ... on PythonError {
      message
      stack
    }
  }
}
"""


class TestGraphs(NonLaunchableGraphQLContextTestMatrix):
    def test_basic_jobs(self, graphql_context):
        result = execute_dagster_graphql(graphql_context, REPOSITORY_QUERY)

        assert result
        assert result.data
        assert result.data["workspaceOrError"]["__typename"] == "Workspace"
        repo_locations = {
            blob["name"]: blob for blob in result.data["workspaceOrError"]["locationEntries"]
        }
        assert "test" in repo_locations
        assert repo_locations["test"]["locationOrLoadError"]["__typename"] == "RepositoryLocation"

        jobs = {
            blob["name"]: blob
            for blob in repo_locations["test"]["locationOrLoadError"]["repositories"][0][
                "pipelines"
            ]
        }

        assert "simple_job_a" in jobs
        assert jobs["simple_job_a"]["graphName"] == "simple_graph"
        assert "simple_job_b" in jobs
        assert jobs["simple_job_b"]["graphName"] == "simple_graph"

    def test_basic_graphs(self, graphql_context, snapshot):
        selector = infer_repository_selector(graphql_context)
        selector.update({"graphName": "simple_graph"})

        result = execute_dagster_graphql(graphql_context, GRAPH_QUERY, {"selector": selector})
        assert result
        assert result.data
        assert result.data["graphOrError"]["__typename"] == "Graph"

        snapshot.assert_match(result.data)



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['TestSensors.test_get_sensor[non_launchable_in_memory_instance_lazy_repository] 1'] = {
    '__typename': 'Sensor',
    'minIntervalSeconds': 30,
    'name': 'always_no_config_sensor',
    'nextTick': None,
    'sensorState': {
        'runs': [
        ],
        'runsCount': 0,
        'status': 'STOPPED',
        'ticks': [
        ]
    },
    'targets': [
        {
            'mode': 'default',
            'pipelineName': 'no_config_pipeline',
            'solidSelection': None
        }
    ]
}

snapshots['TestSensors.test_get_sensor[non_launchable_in_memory_instance_managed_grpc_env] 1'] = {
    '__typename': 'Sensor',
    'minIntervalSeconds': 30,
    'name': 'always_no_config_sensor',
    'nextTick': None,
    'sensorState': {
        'runs': [
        ],
        'runsCount': 0,
        'status': 'STOPPED',
        'ticks': [
        ]
    },
    'targets': [
        {
            'mode': 'default',
            'pipelineName': 'no_config_pipeline',
            'solidSelection': None
        }
    ]
}

snapshots['TestSensors.test_get_sensor[non_launchable_in_memory_instance_multi_location] 1'] = {
    '__typename': 'Sensor',
    'minIntervalSeconds': 30,
    'name': 'always_no_config_sensor',
    'nextTick': None,
    'sensorState': {
        'runs': [
        ],
        'runsCount': 0,
        'status': 'STOPPED',
        'ticks': [
        ]
    },
    'targets': [
        {
            'mode': 'default',
            'pipelineName': 'no_config_pipeline',
            'solidSelection': None
        }
    ]
}

snapshots['TestSensors.test_get_sensor[non_launchable_postgres_instance_lazy_repository] 1'] = {
    '__typename': 'Sensor',
    'minIntervalSeconds': 30,
    'name': 'always_no_config_sensor',
    'nextTick': None,
    'sensorState': {
        'runs': [
        ],
        'runsCount': 0,
        'status': 'STOPPED',
        'ticks': [
        ]
    },
    'targets': [
        {
            'mode': 'default',
            'pipelineName': 'no_config_pipeline',
            'solidSelection': None
        }
    ]
}

snapshots['TestSensors.test_get_sensor[non_launchable_postgres_instance_managed_grpc_env] 1'] = {
    '__typename': 'Sensor',
    'minIntervalSeconds': 30,
    'name': 'always_no_config_sensor',
    'nextTick': None,
    'sensorState': {
        'runs': [
        ],
        'runsCount': 0,
        'status': 'STOPPED',
        'ticks': [
        ]
    },
    'targets': [
        {
            'mode': 'default',
            'pipelineName': 'no_config_pipeline',
            'solidSelection': None
        }
    ]
}

snapshots['TestSensors.test_get_sensor[non_launchable_postgres_instance_multi_location] 1'] = {
    '__typename': 'Sensor',
    'minIntervalSeconds': 30,
    'name': 'always_no_config_sensor',
    'nextTick': None,
    'sensorState': {
        'runs': [
        ],
        'runsCount': 0,
        'status': 'STOPPED',
        'ticks': [
        ]
    },
    'targets': [
        {
            'mode': 'default',
            'pipelineName': 'no_config_pipeline',
            'solidSelection': None
        }
    ]
}

snapshots['TestSensors.test_get_sensor[non_launchable_sqlite_instance_deployed_grpc_env] 1'] = {
    '__typename': 'Sensor',
    'minIntervalSeconds': 30,
    'name': 'always_no_config_sensor',
    'nextTick': None,
    'sensorState': {
        'runs': [
        ],
        'runsCount': 0,
        'status': 'STOPPED',
        'ticks': [
        ]
    },
    'targets': [
        {
            'mode': 'default',
            'pipelineName': 'no_config_pipeline',
            'solidSelection': None
        }
    ]
}

snapshots['TestSensors.test_get_sensor[non_launchable_sqlite_instance_lazy_repository] 1'] = {
    '__typename': 'Sensor',
    'minIntervalSeconds': 30,
    'name': 'always_no_config_sensor',
    'nextTick': None,
    'sensorState': {
        'runs': [
        ],
        'runsCount': 0,
        'status': 'STOPPED',
        'ticks': [
        ]
    },
    'targets': [
        {
            'mode': 'default',
            'pipelineName': 'no_config_pipeline',
            'solidSelection': None
        }
    ]
}

snapshots['TestSensors.test_get_sensor[non_launchable_sqlite_instance_managed_grpc_env] 1'] = {
    '__typename': 'Sensor',
    'minIntervalSeconds': 30,
    'name': 'always_no_config_sensor',
    'nextTick': None,
    'sensorState': {
        'runs': [
        ],
        'runsCount': 0,
        'status': 'STOPPED',
        'ticks': [
        ]
    },
    'targets': [
        {
            'mode': 'default',
            'pipelineName': 'no_config_pipeline',
            'solidSelection': None
        }
    ]
}

snapshots['TestSensors.test_get_sensor[non_launchable_sqlite_instance_multi_location] 1'] = {
    '__typename': 'Sensor',
    'minIntervalSeconds': 30,
    'name': 'always_no_config_sensor',
    'nextTick': None,
    'sensorState': {
        'runs': [
        ],
        'runsCount': 0,
        'status': 'STOPPED',
        'ticks': [
        ]
    },
    'targets': [
        {
            'mode': 'default',
            'pipelineName': 'no_config_pipeline',
            'solidSelection': None
        }
    ]
}

snapshots['TestSensors.test_get_sensors[non_launchable_in_memory_instance_lazy_repository] 1'] = [
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'always_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 60,
        'name': 'custom_interval_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'multi_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'never_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'once_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'running_in_code_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'RUNNING',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    }
]

snapshots['TestSensors.test_get_sensors[non_launchable_in_memory_instance_managed_grpc_env] 1'] = [
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'always_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 60,
        'name': 'custom_interval_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'multi_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'never_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'once_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'running_in_code_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'RUNNING',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    }
]

snapshots['TestSensors.test_get_sensors[non_launchable_in_memory_instance_multi_location] 1'] = [
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'always_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 60,
        'name': 'custom_interval_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'multi_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'never_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'once_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'running_in_code_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'RUNNING',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    }
]

snapshots['TestSensors.test_get_sensors[non_launchable_postgres_instance_lazy_repository] 1'] = [
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'always_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 60,
        'name': 'custom_interval_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'multi_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'never_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'once_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'running_in_code_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'RUNNING',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    }
]

snapshots['TestSensors.test_get_sensors[non_launchable_postgres_instance_managed_grpc_env] 1'] = [
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'always_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 60,
        'name': 'custom_interval_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'multi_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'never_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'once_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'running_in_code_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'RUNNING',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    }
]

snapshots['TestSensors.test_get_sensors[non_launchable_postgres_instance_multi_location] 1'] = [
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'always_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 60,
        'name': 'custom_interval_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'multi_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'never_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'once_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'running_in_code_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'RUNNING',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    }
]

snapshots['TestSensors.test_get_sensors[non_launchable_sqlite_instance_deployed_grpc_env] 1'] = [
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'always_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 60,
        'name': 'custom_interval_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'multi_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'never_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'once_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'running_in_code_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'RUNNING',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    }
]

snapshots['TestSensors.test_get_sensors[non_launchable_sqlite_instance_lazy_repository] 1'] = [
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'always_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 60,
        'name': 'custom_interval_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'multi_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'never_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'once_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'running_in_code_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'RUNNING',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    }
]

snapshots['TestSensors.test_get_sensors[non_launchable_sqlite_instance_managed_grpc_env] 1'] = [
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'always_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 60,
        'name': 'custom_interval_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'multi_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'never_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'once_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'running_in_code_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'RUNNING',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    }
]

snapshots['TestSensors.test_get_sensors[non_launchable_sqlite_instance_multi_location] 1'] = [
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'always_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 60,
        'name': 'custom_interval_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'multi_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'never_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'once_no_config_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'STOPPED',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    },
    {
        'description': None,
        'minIntervalSeconds': 30,
        'name': 'running_in_code_sensor',
        'sensorState': {
            'runs': [
            ],
            'runsCount': 0,
            'status': 'RUNNING',
            'ticks': [
            ]
        },
        'targets': [
            {
                'mode': 'default',
                'pipelineName': 'no_config_pipeline',
                'solidSelection': None
            }
        ]
    }
]



from typing import TYPE_CHECKING, List, Optional, Union

import graphene
from dagster_graphql.implementation.events import iterate_metadata_entries
from dagster_graphql.schema.metadata import GrapheneMetadataEntry
from dagster_graphql.schema.solids import (
    GrapheneCompositeSolidDefinition,
    GrapheneSolidDefinition,
    build_solid_definition,
)

from dagster import AssetKey, check
from dagster.core.host_representation import ExternalRepository, RepositoryLocation
from dagster.core.host_representation.external_data import (
    ExternalAssetNode,
    ExternalStaticPartitionsDefinitionData,
    ExternalTimeWindowPartitionsDefinitionData,
)

from ..implementation.loader import BatchMaterializationLoader
from . import external
from .asset_key import GrapheneAssetKey
from .errors import GrapheneAssetNotFoundError
from .logs.events import GrapheneMaterializationEvent
from .pipelines.pipeline import GrapheneMaterializationCount, GraphenePipeline
from .util import non_null_list

if TYPE_CHECKING:
    from .external import GrapheneRepository


class GrapheneAssetDependency(graphene.ObjectType):
    class Meta:
        name = "AssetDependency"

    asset = graphene.NonNull("dagster_graphql.schema.asset_graph.GrapheneAssetNode")
    inputName = graphene.NonNull(graphene.String)

    def __init__(
        self,
        repository_location: RepositoryLocation,
        external_repository: ExternalRepository,
        input_name: str,
        asset_key: AssetKey,
        materialization_loader: Optional[BatchMaterializationLoader] = None,
    ):
        self._repository_location = check.inst_param(
            repository_location, "repository_location", RepositoryLocation
        )
        self._external_repository = check.inst_param(
            external_repository, "external_repository", ExternalRepository
        )
        self._asset_key = check.inst_param(asset_key, "asset_key", AssetKey)
        self._latest_materialization_loader = check.opt_inst_param(
            materialization_loader, "materialization_loader", BatchMaterializationLoader
        )
        super().__init__(inputName=input_name)

    def resolve_asset(self, _graphene_info):
        asset_node = check.not_none(
            self._external_repository.get_external_asset_node(self._asset_key)
        )
        return GrapheneAssetNode(
            self._repository_location,
            self._external_repository,
            asset_node,
            self._latest_materialization_loader,
        )


class GrapheneAssetNode(graphene.ObjectType):

    # NOTE: properties/resolvers are listed alphabetically
    assetKey = graphene.NonNull(GrapheneAssetKey)
    assetMaterializations = graphene.Field(
        non_null_list(GrapheneMaterializationEvent),
        partitions=graphene.List(graphene.String),
        beforeTimestampMillis=graphene.String(),
        limit=graphene.Int(),
    )
    computeKind: graphene.String
    dependedBy = non_null_list(GrapheneAssetDependency)
    dependedByKeys = non_null_list(GrapheneAssetKey)
    dependencies = non_null_list(GrapheneAssetDependency)
    dependencyKeys = non_null_list(GrapheneAssetKey)
    description = graphene.String()
    id = graphene.NonNull(graphene.ID)
    jobNames = non_null_list(graphene.String)
    jobs = non_null_list(GraphenePipeline)
    latestMaterializationByPartition = graphene.Field(
        graphene.NonNull(graphene.List(GrapheneMaterializationEvent)),
        partitions=graphene.List(graphene.String),
    )
    materializationCountByPartition = non_null_list(GrapheneMaterializationCount)
    metadata_entries = non_null_list(GrapheneMetadataEntry)
    op = graphene.Field(GrapheneSolidDefinition)
    opName = graphene.String()
    partitionKeys = non_null_list(graphene.String)
    partitionDefinition = graphene.String()
    repository = graphene.NonNull(lambda: external.GrapheneRepository)

    class Meta:
        name = "AssetNode"

    def __init__(
        self,
        repository_location: RepositoryLocation,
        external_repository: ExternalRepository,
        external_asset_node: ExternalAssetNode,
        materialization_loader: Optional[BatchMaterializationLoader] = None,
    ):
        self._repository_location = check.inst_param(
            repository_location,
            "repository_location",
            RepositoryLocation,
        )
        self._external_repository = check.inst_param(
            external_repository, "external_repository", ExternalRepository
        )
        self._external_asset_node = check.inst_param(
            external_asset_node, "external_asset_node", ExternalAssetNode
        )
        self._latest_materialization_loader = check.opt_inst_param(
            materialization_loader, "materialization_loader", BatchMaterializationLoader
        )

        super().__init__(
            id=external_asset_node.asset_key.to_string(),
            assetKey=external_asset_node.asset_key,
            opName=external_asset_node.op_name,
            description=external_asset_node.op_description,
        )

    @property
    def repository_location(self) -> RepositoryLocation:
        return self._repository_location

    @property
    def external_repository(self) -> ExternalRepository:
        return self._external_repository

    @property
    def external_asset_node(self) -> ExternalAssetNode:
        return self._external_asset_node

    def get_partition_keys(self) -> List[str]:
        # TODO: Add functionality for dynamic partitions definition
        partitions_def_data = self._external_asset_node.partitions_def_data
        if partitions_def_data:
            if isinstance(
                partitions_def_data, ExternalStaticPartitionsDefinitionData
            ) or isinstance(partitions_def_data, ExternalTimeWindowPartitionsDefinitionData):
                return [
                    partition.name
                    for partition in partitions_def_data.get_partitions_definition().get_partitions()
                ]
        return []

    def resolve_assetMaterializations(
        self, graphene_info, **kwargs
    ) -> List[GrapheneMaterializationEvent]:
        from ..implementation.fetch_assets import get_asset_materializations

        beforeTimestampMillis: Optional[str] = kwargs.get("beforeTimestampMillis")
        try:
            before_timestamp = (
                int(beforeTimestampMillis) / 1000.0 if beforeTimestampMillis else None
            )
        except ValueError:
            before_timestamp = None

        limit = kwargs.get("limit")
        partitions = kwargs.get("partitions")
        if (
            self._latest_materialization_loader
            and limit == 1
            and not partitions
            and not before_timestamp
        ):
            latest_materialization_event = (
                self._latest_materialization_loader.get_latest_materialization_for_asset_key(
                    self._external_asset_node.asset_key
                )
            )

            if not latest_materialization_event:
                return []

            return [GrapheneMaterializationEvent(event=latest_materialization_event)]

        return [
            GrapheneMaterializationEvent(event=event)
            for event in get_asset_materializations(
                graphene_info,
                self._external_asset_node.asset_key,
                partitions,
                before_timestamp=before_timestamp,
                limit=limit,
            )
        ]

    def resolve_computeKind(self, _graphene_info) -> Optional[str]:
        return self._external_asset_node.compute_kind

    def resolve_dependedBy(self, graphene_info) -> List[GrapheneAssetDependency]:
        if not self._external_asset_node.depended_by:
            return []

        materialization_loader = BatchMaterializationLoader(
            instance=graphene_info.context.instance,
            asset_keys=[dep.downstream_asset_key for dep in self._external_asset_node.depended_by],
        )

        return [
            GrapheneAssetDependency(
                repository_location=self._repository_location,
                external_repository=self._external_repository,
                input_name=dep.input_name,
                asset_key=dep.downstream_asset_key,
                materialization_loader=materialization_loader,
            )
            for dep in self._external_asset_node.depended_by
        ]

    def resolve_dependedByKeys(self, _graphene_info) -> List[GrapheneAssetKey]:
        return [
            GrapheneAssetKey(path=dep.downstream_asset_key.path)
            for dep in self._external_asset_node.depended_by
        ]

    def resolve_dependencyKeys(self, _graphene_info):
        return [
            GrapheneAssetKey(path=dep.upstream_asset_key.path)
            for dep in self._external_asset_node.dependencies
        ]

    def resolve_dependencies(self, graphene_info) -> List[GrapheneAssetDependency]:
        if not self._external_asset_node.dependencies:
            return []

        materialization_loader = BatchMaterializationLoader(
            instance=graphene_info.context.instance,
            asset_keys=[dep.upstream_asset_key for dep in self._external_asset_node.dependencies],
        )
        return [
            GrapheneAssetDependency(
                repository_location=self._repository_location,
                external_repository=self._external_repository,
                input_name=dep.input_name,
                asset_key=dep.upstream_asset_key,
                materialization_loader=materialization_loader,
            )
            for dep in self._external_asset_node.dependencies
        ]

    def resolve_jobNames(self, _graphene_info) -> List[str]:
        return self._external_asset_node.job_names

    def resolve_jobs(self, _graphene_info) -> List[GraphenePipeline]:
        job_names = self._external_asset_node.job_names or []
        return [
            GraphenePipeline(self._external_repository.get_full_external_pipeline(job_name))
            for job_name in job_names
            if self._external_repository.has_external_pipeline(job_name)
        ]

    def resolve_latestMaterializationByPartition(
        self, graphene_info, **kwargs
    ) -> List[Optional[GrapheneMaterializationEvent]]:
        from ..implementation.fetch_assets import get_asset_materializations

        get_partition = (
            lambda event: event.dagster_event.step_materialization_data.materialization.partition
        )

        partitions = kwargs.get("partitions") or self.get_partition_keys()

        events_for_partitions = get_asset_materializations(
            graphene_info,
            self._external_asset_node.asset_key,
            partitions,
        )

        latest_materialization_by_partition = {}
        for event in events_for_partitions:  # events are sorted in order of newest to oldest
            event_partition = get_partition(event)
            if event_partition not in latest_materialization_by_partition:
                latest_materialization_by_partition[event_partition] = event
            if len(latest_materialization_by_partition) == len(partitions):
                break

        # return materializations in the same order as the provided partitions, None if
        # materialization does not exist
        ordered_materializations = [
            latest_materialization_by_partition.get(partition) for partition in partitions
        ]

        return [
            GrapheneMaterializationEvent(event=event) if event else None
            for event in ordered_materializations
        ]

    def resolve_materializationCountByPartition(
        self, graphene_info
    ) -> List[GrapheneMaterializationCount]:
        asset_key = self._external_asset_node.asset_key
        partition_keys = self.get_partition_keys()

        count_by_partition = graphene_info.context.instance.get_materialization_count_by_partition(
            [self._external_asset_node.asset_key]
        )[asset_key]

        return [
            GrapheneMaterializationCount(partition_key, count_by_partition.get(partition_key, 0))
            for partition_key in partition_keys
        ]

    def resolve_metadata_entries(self, _graphene_info) -> List[GrapheneMetadataEntry]:
        return list(iterate_metadata_entries(self._external_asset_node.metadata_entries))

    def resolve_op(
        self, _graphene_info
    ) -> Optional[Union[GrapheneSolidDefinition, GrapheneCompositeSolidDefinition]]:
        if len(self._external_asset_node.job_names) >= 1:
            pipeline_name = self._external_asset_node.job_names[0]
            pipeline = self._external_repository.get_full_external_pipeline(pipeline_name)
            return build_solid_definition(pipeline, self._external_asset_node.op_name)
        else:
            return None

    def resolve_partitionDefinition(self, _graphene_info) -> Optional[str]:
        partitions_def_data = self._external_asset_node.partitions_def_data
        if partitions_def_data:
            return str(partitions_def_data.get_partitions_definition())
        return None

    def resolve_partitionKeys(self, _graphene_info) -> List[str]:
        return self.get_partition_keys()

    def resolve_repository(self, graphene_info) -> "GrapheneRepository":
        return external.GrapheneRepository(
            graphene_info.context.instance, self._external_repository, self._repository_location
        )


class GrapheneAssetNodeOrError(graphene.Union):
    class Meta:
        types = (GrapheneAssetNode, GrapheneAssetNotFoundError)
        name = "AssetNodeOrError"



from typing import Any

from dagster_graphql.test.utils import execute_dagster_graphql

from .graphql_context_test_suite import GraphQLContextVariant, make_graphql_context_test_suite

INSTANCE_QUERY = """
query InstanceDetailSummaryQuery {
    instance {
        runQueuingSupported
        hasInfo
    }
}
"""


BaseTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[
        GraphQLContextVariant.sqlite_with_queued_run_coordinator_managed_grpc_env(),
    ]
)


class TestInstanceSettings(BaseTestSuite):
    def test_instance_settings(self, graphql_context):
        results = execute_dagster_graphql(graphql_context, INSTANCE_QUERY)
        assert results.data == {"instance": {"runQueuingSupported": True, "hasInfo": True}}
